// Copyright 2022 Clayton Craft <clayton@craftyguy.net>
// SPDX-License-Identifier: GPL-3.0-or-later

package service

import (
	"errors"
	"fmt"
	"os"
	"strings"
	"sync"
	"time"

	goCmd "git.sr.ht/~craftyguy/go-cmd"
	"sr.ht/~craftyguy/superd/internal/cmd"
	"sr.ht/~craftyguy/superd/internal/config"
	"sr.ht/~craftyguy/superd/internal/env"
	"sr.ht/~craftyguy/superd/internal/log"
	"sr.ht/~craftyguy/superd/pkg/api"
)

type Service struct {
	Start   chan bool
	Stop    chan bool
	Details chan api.ServiceDetails

	reload chan bool
	logger *log.Logger
	// make cmd.NewCommand overridable in testing
	newCommand func(exe string, args []string) cmd.Commander

	// func for one-shot getting command status
	getCmdStatus func() goCmd.Status

	before struct {
		list []string
		sync.RWMutex
	}
	after struct {
		list []string
		sync.RWMutex
	}

	name struct {
		value string
		sync.RWMutex
	}

	config struct {
		value config.ServiceConfig
		sync.RWMutex
	}
}

// whether or not the service should be autostarted
func (s *Service) Autostart() bool {
	return s.Config().Autostart
}

func (s *Service) Name() string {
	s.name.RLock()
	defer s.name.RUnlock()
	return s.name.value
}

func (s *Service) AddAfter(name string) {
	s.after.Lock()
	defer s.after.Unlock()
	for _, n := range s.after.list {
		if n == name {
			return
		}
	}
	s.after.list = append(s.after.list, name)
}

func (s *Service) AddBefore(name string) {
	s.before.Lock()
	defer s.before.Unlock()
	for _, n := range s.before.list {
		if n == name {
			return
		}
	}
	s.before.list = append(s.before.list, name)
}

func (s *Service) After() []string {
	s.after.RLock()
	defer s.after.RUnlock()
	// return a copy
	return append([]string{}, s.after.list...)
}

func (s *Service) Before() []string {
	s.before.RLock()
	defer s.before.RUnlock()
	// return a copy
	return append([]string{}, s.before.list...)
}

func (s *Service) SetConfig(c config.ServiceConfig) {
	s.config.Lock()
	defer s.config.Unlock()
	s.config.value = c
	go func() {
		select {
		case <-time.After(1 * time.Second):
			s.logger.Warnln(s.Name(), "timed out trying to signal config reload, this shouldn't have happened....")
		case s.reload <- true:
		}
	}()
}

func (s *Service) Config() config.ServiceConfig {
	s.config.RLock()
	defer s.config.RUnlock()
	return s.config.value
}

func Init(name string) (s *Service) {

	s = &Service{
		Start:   make(chan bool),
		Stop:    make(chan bool),
		Details: make(chan api.ServiceDetails),
		reload:  make(chan bool),
		logger:  log.Get(),
	}
	s.name.Lock()
	defer s.name.Unlock()
	s.name.value = name
	// make cmd.NewCommand overridable in testing
	s.newCommand = cmd.NewCommand

	return
}

func (s *Service) runCmd(c cmd.Commander) (err error) {

	// environment vars are collected from various sources in order of
	// importance, from least to most, allowing later sources (e.g. service
	// config file) to override previous var values
	eMap := make(map[string]string)

	// add "global" environment
	for v := range env.Init().IterVars() {
		eMap[v[0]] = v[1]
	}

	// import all variables from current env
	for _, v := range os.Environ() {
		name, value, _ := strings.Cut(v, "=")
		eMap[name] = value
	}

	// set service-specific env. from config
	for k, v := range s.Config().Env {
		eMap[k] = v
	}

	// format env into a list
	var e []string
	for k, v := range eMap {
		e = append(e, fmt.Sprintf("%s=%s", k, v))
	}

	switch s.Config().Type {
	case config.TypeSimple:
		err = c.Start(e)
	case config.TypeOneshot:
		err = c.Start(e)
	case config.TypeNotify:
		err = c.StartNotify(e, s.Name()+"_notify.sock", s.Config().TimeoutStartDur)
	default:
		err = fmt.Errorf("unknown type: %s", s.Config().Type)
	}
	return
}

// startup service routine, e.g. running ExecStartPre, ExecStart. Returns
// channels for receiving the command object if successful or service status on
// failure. the cancel chan is used to signal to the routine to abort startup (e.g. when retrying)
func (s *Service) startup(cancel <-chan struct{}) (cmdChan chan cmd.Commander, startStatus chan api.ServiceStatus) {
	startStatus = make(chan api.ServiceStatus)
	cmdChan = make(chan cmd.Commander)

	go func() {
		defer close(startStatus)
		defer close(cmdChan)

		conf := s.Config()

		// ExecStartPre
		if conf.ExecStartPreCmd != "" {
			preCmd := s.newCommand(conf.ExecStartPreCmd, conf.ExecStartPreCmdArgs)
			if err := s.runCmd(preCmd); err != nil && !errors.As(err, &cmd.ErrCmdExited{}) {
				startStatus <- api.ServiceFailed
				return
			}
		}

		var startCmd cmd.Commander
		// ExecStart

		// burst limit timing
		limitStart := time.Now()
		tries := 0

	start_loop:
		for {
			// if retry/burst limit is exceeded during the start limit duration, then fail
			if tries >= conf.StartLimitBurst {
				if time.Since(limitStart) < conf.StartLimitDur {
					s.logger.Errorf(s.Name(), "failed to start service: exceeded the max number of tries, %d, in %s",
						conf.StartLimitBurst, conf.StartLimitDur.String())
					startStatus <- api.ServiceFailed
					return
				}

				// burst limit wasn't exceeded over the burst duration, so reset and keep going
				limitStart = time.Now()
				tries = 0
			}

			startCmd = s.newCommand(conf.Cmd, conf.CmdArgs)
			if err := s.runCmd(startCmd); err != nil {
				switch err.(type) {
				case cmd.ErrCmdExitedNonZero:
				case cmd.ErrUnhandledStatus:
				case cmd.ErrCmdExited:
					// oneshot is expected to exit
					if conf.Type == config.TypeOneshot {
						break start_loop
					}
				case cmd.ErrTimeout:
				default:
					s.logger.Errorln(s.Name(), "unknown failure when trying to start service: "+err.Error())
				}
				s.logger.Errorln(s.Name(), "failed to start service: "+err.Error())
			} else {
				break
			}

			// either the restart timer expires, or a signal is received to
			// cancel the startup
			select {
			case <-time.After(conf.RestartDur):
				tries++
				s.logger.Errorln(s.Name(), "retrying...")
				continue
			case <-cancel:
				s.logger.Debugln(s.Name(), "service startup canceled")
				return
			}
		}

		// ExecStartPost
		if conf.ExecStartPostCmd != "" {
			postCmd := s.newCommand(conf.ExecStartPostCmd, conf.ExecStartPostCmdArgs)
			if err := s.runCmd(postCmd); err != nil && !errors.As(err, &cmd.ErrCmdExited{}) {
				startStatus <- api.ServiceFailed
				return
			}
		}

		// send the cmd from ExecStart
		cmdChan <- startCmd
	}()

	return
}

func (s *Service) Uptime() int64 {
	if s.getCmdStatus == nil {
		return 0
	}
	return int64(s.getCmdStatus().Runtime)
}

// make killTimeout overridable in testing
var killTimeout = time.Duration(5 * time.Second)

func (s *Service) Run(quit <-chan struct{}) {

	var (
		startStatus    <-chan api.ServiceStatus
		onFailComplete chan struct{}
		restartTimer   <-chan time.Time
		details        api.ServiceDetails
		// used for signaling that the service should *not* be restarted, e.g.
		// it was explicitly stopped
		shouldRestart bool
		// used for signaling that the service was explicitly stopped, so that
		// status can be reflected accurately
		shouldStop bool

		// used for receiving the started service process details
		cmdChan         <-chan cmd.Commander
		command         cmd.Commander
		cancelStartChan chan struct{}

		// command streams
		cmdStatus <-chan goCmd.Status
		cmdStdout <-chan string
		cmdStderr <-chan string
	)
	details.Status = api.ServiceStopped

	for {
		select {

		// Start (process) requested
		case <-s.Start:
			if details.Status != api.ServiceStarted && details.Status != api.ServiceStarting {
				shouldRestart = true
				shouldStop = false
				cancelStartChan = make(chan struct{})
				cmdChan, startStatus = s.startup(cancelStartChan)
				details.Status = api.ServiceStarting
			}

		// new service process info received (e.g. from startup())
		case command = <-cmdChan:
			if command == nil {
				// startup was canceled
				details.Status = api.ServiceStopped
				continue
			} else {
				close(cancelStartChan)
				cancelStartChan = nil
			}
			startStatus = nil
			cmdChan = nil
			streams := command.Streams()
			cmdStatus = streams.Status
			cmdStdout = streams.Stdout
			cmdStderr = streams.Stderr
			details.Status = api.ServiceStarted
			details.Pid = command.Status().PID
			s.getCmdStatus = command.Status

		// service startup status
		case ss := <-startStatus:
			startStatus = nil
			cmdStatus = nil
			cmdChan = nil
			switch ss {
			case api.ServiceFailed:
				details.Status = api.ServiceFailed
				// Run execOnFailure
				if s.Config().ExecOnFailureCmd != "" {
					onFailComplete = make(chan struct{})
					s.execOnFail(onFailComplete)
				}
				// Note: retries have already happened, so at this point the
				// service is "failed"
			}

		// Stop (process) requested
		case <-s.Stop:
			// Cancel restart and start timer in so that it doesn't trigger and
			// restart the service
			restartTimer = nil

			// don't allow the service to be restarted
			shouldRestart = false

			// indicate that the service is stopped because it was told to
			shouldStop = true

			if cancelStartChan != nil {
				close(cancelStartChan)
				cancelStartChan = nil
			}

			if onFailComplete != nil {
				close(onFailComplete)
				onFailComplete = nil
			}

			// guard against many rapid calls to stop
			if command != nil {
				if s.Config().Type == config.TypeOneshot {
					// oneshot services can't be "stopped", there should be no
					// process runninng... so just indicate that the status is
					// "stopped" when requested to
					details.Status = api.ServiceStopped
				} else if err := command.Stop(killTimeout); err != nil {
					s.logger.Errorln(s.Name(), "failed to stop: "+err.Error())
					details.Status = api.ServiceFailed
				}
				// for other situations, the service's status is set when the
				// updated cmdStatus is received later
				command = nil
			}

		// Quit (goroutine) requested
		case <-quit:
			// Note: no cleanup happens here, so it's expected that the manager
			// will have signaled "Stop" before this channel is closed
			return

		// Send details/status
		case s.Details <- details:

		// Status from running process
		case status := <-cmdStatus:

			// stopped
			if status.StopTs > 0 {
				s.logger.Debugln(s.Name(), "exited with: "+fmt.Sprint(status.Exit))
				s.getCmdStatus = nil

				failed := false
				if status.Error != nil {
					failed = true
				} else if status.Exit != 0 {
					// err from os/exec
					failed = true
				} else if !status.Complete {
					// command didn't run to completion, it was stopped(?)
					// note: setting failed not strictly necessary here, but
					// it's readable.
					failed = false
				} else {
					// command completed successfully
					if s.Config().Type == config.TypeOneshot {
						// show 'started' for oneshot services
						details.Status = api.ServiceStarted
					} else {
						details.Status = api.ServiceStopped
					}
				}

				if failed {
					if shouldStop {
						// explicitly stopped services shouldn't have a 'failed'
						// status. This can happen when an app receives a SIGTERM
						// but doesn't handle it, and exits with something like -1.
						// Technically the app "failed" to run to completion, but
						// for our purposes it did stop successfully (i.e. it's not
						// running anymore).
						details.Status = api.ServiceStopped
						continue
					}

					details.Status = api.ServiceFailed

					// fire off any execOnFailure
					if s.Config().ExecOnFailureCmd != "" {
						onFailComplete = make(chan struct{})
						s.execOnFail(onFailComplete)

						// don't schedule a restart until after the
						// ExecOnFailure command has completed
						continue
					}
				} else {
					details.Status = api.ServiceStopped
				}

				if shouldRestart {
					restartTimer = s.scheduleRestart(details.Status)
				}

			} else {
				s.logger.Warnln(s.Name(), "received unhandled status: "+fmt.Sprint(status))
			}

		// onFailStatus -- from ExecOnFailure
		case <-onFailComplete:
			onFailComplete = nil
			// Restart scheduled if necessary
			restartTimer = s.scheduleRestart(details.Status)

		// Restart timer
		case <-restartTimer:
			go func() {
				select {
				case <-time.After(1 * time.Second):
					s.logger.Warnln(s.Name(), "timed out trying to start the service, this shouldn't have happened....")
				case s.Start <- true:
				}
			}()

		// Log output from command
		case line, open := <-cmdStderr:
			if !open {
				cmdStderr = nil
			} else {
				s.logger.Println(s.Name(), line)
			}
		case line, open := <-cmdStdout:
			if !open {
				cmdStdout = nil
			} else {
				s.logger.Println(s.Name(), line)
			}

		// config should be reloaded
		case <-s.reload:
			// update details. note that this is only updating fields that
			// aren't changed by starting/stopping the service, or other
			// runtime events related to the supervised service.
			details.Autostart = s.Autostart()
			details.ConfigPath = s.Config().Path
			details.Type = string(s.Config().Type)
			details.Description = s.Config().Description
		}
	}
}

func (s *Service) oneshotShouldRestart(status api.ServiceStatus) bool {
	// oneshot should only restart if it is set to Always or OnFailure, and
	// only when the status is "failed"
	if status == api.ServiceFailed {
		if s.Config().Restart == config.RestartAlways || s.Config().Restart == config.RestartFailure {
			return true
		}
	}
	return false
}

func (s *Service) simpleShouldRestart(status api.ServiceStatus) bool {
	if status == api.ServiceFailed {
		if s.Config().Restart == config.RestartAlways || s.Config().Restart == config.RestartFailure {
			return true
		}
	} else if status == api.ServiceStopped {
		if s.Config().Restart == config.RestartAlways {
			return true
		}
	}

	return false
}

// Checks whether a restart should be scheduled based on the given status and
// the service's config, and returns a channel to a timer for when it should be
// restarted. The "timer" returned will have been started.
func (s *Service) scheduleRestart(status api.ServiceStatus) (timer <-chan time.Time) {
	var restart bool

	switch s.Config().Type {
	case config.TypeOneshot:
		restart = s.oneshotShouldRestart(status)
	case config.TypeSimple:
		restart = s.simpleShouldRestart(status)
	default:
		s.logger.Warnf(s.Name(), "unable to determine restart behavior, unknown service type: %s", s.Config().Type)
	}

	if restart {
		s.logger.Debugln(s.Name(), fmt.Sprintf("restarting in %d second(s)", uint64(s.Config().RestartDur.Seconds())))
		timer = time.After(s.Config().RestartDur)
		return
	}
	return
}

// Run the ExecOnFailure cmd and close the given channel when done
func (s *Service) execOnFail(done chan struct{}) {
	c := s.newCommand(s.Config().ExecOnFailureCmd, s.Config().ExecOnFailureCmdArgs)
	if err := s.runCmd(c); err != nil && !errors.As(err, &cmd.ErrCmdExited{}) {
		s.logger.Errorf(s.Name(), "failed to run ExecOnFailure %q: %s\n", s.Config().ExecOnFailureCmd, err)
	}
	close(done)
}
